/****************************************************************************
 * Игра "полицейские и грабитель" (дипломная работа, мат-мех СПбГУ)         *
 * @author Пётр Пейсаховский                                                *
 * @date 10.05.2016                                                         *
 ****************************************************************************/
function getInfinity() {return 10000000}

/**
 *  Находит максимальный элемент в массиве
 *  Пример: N = [1,2,3].max() 
 */
Array.prototype.max = function() {
    return Math.max.apply(null, this);
};

/**
 * Находит минимальный элемент в массиве
 */
Array.prototype.min = function() {
    return Math.min.apply(null, this);
};

/** 
 * Находит прямое проихведение массивов
 * Пример: Prod([1,2],[3,4]) = [[1,3],[1,4],[2,3],[2,4]] 
 */
function Prod(paramArray) {
    function addTo(curr, args) {
        var i, copy, 
            rest = args.slice(1),
            last = !rest.length,
            result = [];
        for (i = 0; i < args[0].length; i++) {
            copy = curr.slice();
            copy.push(args[0][i]);
            if (last) {
                result.push(copy);
            } else {
                result = result.concat(addTo(copy, rest));
            }
        }
        return result;
    }
  
    if (Array.isArray(paramArray[0]))
        return addTo([], Array.prototype.slice.call(paramArray));
    else    
        return addTo([], Array.prototype.slice.call(arguments));
}

/**
 *  Генерирует квадратную матрицу заполненную нулями
 */
function GetNullMatrix(n){
    var M = [];
    for (var i=0; i<n; i++){
        var L = [];
        for (var j=0; j<n; j++){            
            L.push(0);
        }
        M.push(L);
    }
    return M;
}

/**
 *  "Добавляет" ребро в матрице инцедентности 
 */
function Connect(M,i,j){
    M[i][j]=1; M[j][i]=1;
}


/**
 *  Считает корень n-ой степени
 */
function SqrtN(X,n){
    return Math.exp( Math.log(X) / n );
}


/**
 *  Генератор графов примеров.
 *  Может генерировать матрицу инцидентности для:
 *    тора, куба, n-мерного куба, сетки, цикла, дерева и т.п.  
 */
GraphGenerator = {
    // Сетка на r-мерном кубе
    nCube: function(M, k, GraphParams){
    var n = M.length;    

    //if (!(r>0)) r = 3;
    var r = GraphParams.Dimensions;

    var size = Math.floor(SqrtN(n, r));
    Coords = [];
    for (var i=0; i< Math.pow(size, r); i++){        

        var coords = [];
        var N = i;
        for (p=r; p>=1; p--)
        {
            var C = Math.floor(N / Math.pow(size, p-1));
            coords.push(C);
            N = N - C * Math.pow(size, p-1);
        }
        Coords.push(coords);
        for (p=0; p<=r-1; p++){
            if (coords[r-1-p]<size-1) Connect(M, i, i+Math.pow(size, p))
            //if (coords[r-1-p]==0) { Connect(M,i, i + (size-1)*Math.pow(size, p)); }
        }
    }
    return M;
  },


  // Сетка на кубе
  cube: function(M,k){
    var n = M.length;    
    var size = Math.floor(SqrtN(n,3));

    for (var i=0; i<size*size*size; i++){        
        var z=Math.floor(i / size / size);        
        var y=Math.floor((i-z*size*size) / size);
        var x=i-z*size*size-y*size;        

        if (x<size-1) Connect(M,i,i+1);
        if (y<size-1) Connect(M,i,i+size);
        if (z<size-1) Connect(M,i,i+size*size);

        /*if (x==0) { Connect(M,i, i + size -1); }        
        if (y==0) { Connect(M,i, i + size * (size - 1)); }
        if (z==0) { Connect(M,i, i + size * size * (size - 1));}*/
    }
    return M;
  },

  // Сетка на торе
  torus: function(M,k){
    var n = M.length;    
    var size = math.floor(math.sqrt(n));
    for (var i=0; i<size*size; i++){
        var x=math.mod(i, size);
        var y=math.floor(i / size);
        if (x<size-1) {M[i][i+1]=1; M[i+1][i]=1;}    
        if (y<size-1) {M[i][i+size]=1; M[i+size][i]=1;}           
    }

    for (var i=0; i<size; i++){
        M[i*size][i*size+size-1] = 1;
        M[i*size+size-1][i*size] = 1;

        M[i][i+size*(size-1)] = 1;
        M[i+size*(size-1)][i] = 1;


    }
    return M;
  },

  // Сетка на цилиндре
  cylinder: function(M,k){
    var n = M.length;    
    var size = math.floor(math.sqrt(n));
    for (var i=0; i<size*size; i++){
        var x=math.mod(i, size);
        var y=math.floor(i / size);
        if (x<size-1) {M[i][i+1]=1; M[i+1][i]=1;}    
        if (y<size-1) {M[i][i+size]=1; M[i+size][i]=1;}           
    }

    // Замыкаем тор по краям
    for (var i=0; i<size; i++){
        M[i*size][i*size+size-1] = 1;
        M[i*size+size-1][i*size] = 1;
    }

    return M;
  },

  // Сетка
  grid: function(M,k){
    var n = M.length;    
    var size = math.floor(math.sqrt(n));
    for (var i=0; i<size*size; i++){
        var x=math.mod(i, size);
        var y=math.floor(i / size);
        if (x<size-1) {M[i][i+1]=1; M[i+1][i]=1;}    
        if (y<size-1) {M[i][i+size]=1; M[i+size][i]=1;}   
    }
    return M;
  },

  // Цикл
  cycle: function(M,k){
    var n = M.length;    
    for (var i=0; i<n; i++){

        j = i+1;
        if (j>=n) j = 0;

        M[i][j]=1;
        M[j][i]=1;
    }
    return M;
  },

  // Дерево
  tree: function(M,k){
    var n = M.length;    
    for (var i=0; i<n-1; i++){
        j = i;
        while (i>=j)
            j = Math.floor(Math.random()*n);

        M[i][j]=1;
        M[j][i]=1;
    }
    return M;
  },


  // Случайный граф
  random: function(M,k){
    var n = M.length;
    // Ставим единичку в каждой строчке
    for (var i=0; i<n; i++){
        j = i;
        while (i==j)
            j = Math.floor(Math.random()*n);
        M[i][j]=1;
           M[j][i]=1;
    }


    // И добавляем дополнительных соединений
    for (var x=0; x<n; x++)
    	for (var y=0; y<n; y++)
	    	if (Math.random()<k)
	    	{    	
		        M[x][y] = 1;
	        	M[y][x] = 1;
	    	}
    return M;
  }

}

/**
 *  Создает список вершин с метками для визуализации
 */
function generateNodes(n){
    var L = [];
    for (var i=0; i<n; i++){
        L.push({id: i, label: i})
    }
    return L;
}

/**
 *  Создаёт список рёбер для визуализации
 */
function generateEdges(M){
    var n = M.length;
    var L = [];
    for (var i=0; i<n; i++){    
        for (var j=0; j<n; j++){            
            if (i>j)
                if (M[i][j] > 0)
                    L.push({from: i, to: j});
        }       
    }   
    return L;
}

/**
 *  Возвращает список вершин соединённых с указанной вершиной ребром
 */
function getLinkedNodes(M, k, selfConnected){
    var n = M.length;
    var L = [];
    for (var i=0; i<n; i++){
        if (M[k][i] > 0)
            L.push(i);
    }

    if ((selfConnected) && (L.length == 0))
        L.push(k);

    return L;
}


/**
 *  Объект - игра C&R
 */
Game={
    CopsPos : [],
    CopsAmount : 3,
    Robber : 5,
    CopToMove : 0,
    Matrix: [],
    GameOver: false,
    SavedState: {},
    GraphParams: {NodesAmount: 27, Dimensions: 3, GraphType: "cube"}, // Параметры графа по умолчанию

    alert: function (a,b,c){
        if (typeof process !== 'undefined' ){
            process.stdout.write("["+a+"]\n");
            //process.stdout.write("["+a+"]\n");
        }
    },

    // Для отладки
    debug_alert: function (a,b,c){
        if (typeof process !== 'undefined' ){
            process.stdout.write("["+a+"]\n");
            //process.stdout.write("["+a+"]\n");
        }
    },

    /**
     * Сохраняет текущее состояние игры
     */
    SaveState: function (){
        Game.SavedState.CopsPos    = Game.CopsPos.slice();
        Game.SavedState.CopsAmount = Game.CopsAmount;
        Game.SavedState.Robber     = Game.Robber;
        Game.SavedState.CopToMove  = Game.CopToMove;
        Game.SavedState.GameOver   = Game.GameOver;
        Game.SavedState.pitfalls   = Game.pitfalls;
    },

    /**
     * Восстанавливает сохранённое состояние игры
     */
    LoadState: function (){
        Game.CopsPos    = Game.SavedState.CopsPos.slice();
        Game.CopsAmount = Game.SavedState.CopsAmount;
        Game.Robber     = Game.SavedState.Robber;
        Game.CopToMove  = Game.SavedState.CopToMove;
        Game.GameOver   = Game.SavedState.GameOver;
        Game.pitfalls   = Game.SavedState.pitfalls;
    },

    /**
     *  Вычисляет ход для копов по сопряженному алгоритму
     */
    getCopsMove: function(){
        Game.SaveState()
        try{
            //var c1 = getLinkedNodes(Game.Matrix, Game.CopsPos[0]);
            //var c2 = getLinkedNodes(Game.Matrix, Game.CopsPos[1]);
            //Game.PossibleMoves =  Prod(c1,c2);
            var C;
    
            if (Game.CopsPos.indexOf(-1)>=0){                
                var range = Array.apply(null, {length: Game.Matrix.length}).map(Number.call, Number);
                C = Array.apply(null, {length: Game.CopsAmount}).map(Function.call, function() {return range});
            }

            else
            {
                C = Game.CopsPos.map(
                    function (pos) {return getLinkedNodes(Game.Matrix, pos, true);}
                )
                Game.debug_alert("Link list built.");
            }


            Game.PossibleMoves = Prod(C);
            Game.debug_alert("Production calculated: " + Game.PossibleMoves.length);
    
            // Делаем все возможные ходы. Выбираем ход с наихудшим последующим ходом грабителя        
            var BestCopMove = {Robber:{potential:getInfinity()+1}};        
            for (var i = 0; i < Game.PossibleMoves.length; i++) 
            {
                Game.LoadState();
                Game.CopsPos = Game.PossibleMoves[i]; //[Game.PossibleMoves[i][0], Game.PossibleMoves[i][1]];
                R = Game.getRobberMove();
                Game.PossibleMoves[i].Robber = R;
      
                if (Game.Robber == Game.PossibleMoves[i][0]){
                    BestCopMove = Game.PossibleMoves[i];
                    return BestCopMove;//[BestCopMove[0], BestCopMove[1]];
                }
    
                if (Game.Robber == Game.PossibleMoves[i][1]){
                    BestCopMove = Game.PossibleMoves[i];
                    return BestCopMove; //[BestCopMove[0], BestCopMove[1]];
                }
    
                if (Game.PossibleMoves[i].Robber.potential < BestCopMove.Robber.potential){
                    BestCopMove = Game.PossibleMoves[i];
                }
    
            }
            Game.BestCopMove =BestCopMove;
            return BestCopMove;//[BestCopMove[0], BestCopMove[1]];
        }
        finally
        {
            Game.LoadState()
  
        }
    },

    /**
     *  Делает автоматический ход копами (по сопряженному алгоритму)
     */
    AutoMoveCops: function(){
    	var InitMove = Game.CopsPos.indexOf(-1)>-1;
        if (InitMove) Game.debug_alert("Making first move");

        Game.CopsPos = Game.getCopsMove();
        Game.CopToMove = 0;
  
        if (InitMove){
        	Game.initPitfalls();
    	}

        if (InitMove) Game.debug_alert("first move done.");

        if (Game.IsOnCop(Game.Robber)){
            Game.GameOver = true;
            Game.alert("Ура! Вы поймали грабителя!", "Победа", "robber.jpg");
            return true;
        }  
        else           
            Game.DelayedMoveRobber();
    },  

    /**
     *  Передвигает копа в указанную вершину
     *  (используется для ручных ходов копами)
     */
    MoveCop: function (node){
    	var InitMove = Game.CopsPos.indexOf(-1)>-1;
        Game.CopsPos[Game.CopToMove] = node;
        Game.CopToMove++;

        if (Game.IsOnCop(Game.Robber)){
            Game.GameOver = true;
            Game.alert("Ура! Вы поймали грабителя!", "Победа", "robber.jpg");
        }

        if (Game.CopToMove>=Game.CopsAmount){
            Game.CopToMove = 0;
            Game.DelayedMoveRobber();
        }
        if (InitMove){
        	Game.initPitfalls();
    	}
    },

    /**
     *  Проверяет, что в узле есть коп
     */
    IsOnCop: function(r){
        return Game.CopsPos.indexOf(r) > -1;
    },

    /**
     *  Вычисляет ход для грабителя потенциальным алгоритмом
     */
    getRobberMove: function(){
        if (Game.Robber < 0){
            var MP = Game.getNodeWithMaxPotential();
            return {node: MP.index, potential: MP.potential};
        }

        var Nodes = getLinkedNodes(Game.Matrix, Game.Robber);
        Nodes.push(Game.Robber);

        var Ps = Game.calcPotetials();

        var NodeToRun = 0;
        var NodePotential = -1;

        for (var i=0; i<Game.nodesNo; i++){
            if (Ps[Nodes[i]] > NodePotential){
                NodeToRun = Nodes[i];
                NodePotential = Ps[Nodes[i]];
            }
        }   

        return {node:NodeToRun, potential: NodePotential};      
    },

    DelayedMoveRobber: function(){
      Game.MoveRobber();
      //setTimeout(Game.MoveRobber, 500);
      //setTimeout(unblink, 600);
    },

    /**
     *  Делает ход грабителем
     */
    MoveRobber: function (){
        if (Game.GameOver)
            return;

        var NodeToRun = Game.getRobberMove().node;


        if ((Game.Robber > -1) && (Game.IsOnCop(NodeToRun) || (NodeToRun == Game.Robber))) {
            Game.alert("Грабитель пропускает ход");
        }
        else
        Game.Robber = NodeToRun;

    },

    /**
     *  Инициализации игры
     */
    init: function(N){
        Game.debug_alert("init");
        Game.GameOver = false;            

        var nodesNo0; 
        if (N) nodesNo0 = N; else nodesNo0 = Game.GraphParams.NodesAmount; //125;//15+Math.floor(Math.random()*20);

        var M0 = Game.GenerateGraph(GetNullMatrix(nodesNo0), 0.025, Game.GraphParams);


        Game.debug_alert("matrix generated");

        Game.Matrix = M0;
        Game.nodesNo = nodesNo0;
        var M = Game.Matrix;
        A = M;

        Game.Ms = [M];
        
        for (var i=1; i<=Game.nodesNo; i++){
            A = math.multiply(A, M);
            Game.Ms.push(A);
        }   

        Game.debug_alert("Ms generated");

        Game.initPitfalls();

        Game.debug_alert("pitfalls generated");

        Game.Robber = 1;
        var Ps = Game.calcInitPotetials();
        var n1 = Ps.indexOf(Ps.max());
        Ps[n1] = -1;
        var n2 = Ps.indexOf(Ps.max());
        Game.CopsPos = [n1,n2];


        Game.GraphLinkedNodes = [];
        for (var i=0; i<Game.Matrix.length; i++) 
            Game.GraphLinkedNodes.push(getLinkedNodes(Game.Matrix,i));


        Game.debug_alert("Initialization done");

    },

    /**
     *  Находит расстояние между парой вершин
     */
    getDistance: function (k,l){
        if (k==l)
            return 0;

        for (var i=0; i<Game.nodesNo; i++){
            if (Game.Ms[i][k][l] > 0)
                return i+1;
        }
        return getInfinity();
    },

    /**
     *  Находит расстояние от вершины до ближайшего копа
     */
    getDistanceToCops: function(i){
        var dists = [];
        for (var j=0; j<Game.CopsAmount; j++){
            dists.push( Game.getDistance(i, Game.CopsPos[j]) );
        }
        return dists.min();
    },

    /**
     *  Находит вершину с максимальным потенциалом
     */
    getNodeWithMaxPotential: function(){
        var Ps = Game.calcPotetials();
        var ind = Ps.indexOf(Ps.max());
        var P = Ps.max();
        return {index: ind, potential: P};
    },
    
    /**
     *  Расчитывает начальные потенциалы вершин (расстояния до ближайших копов)
     */
    calcInitPotetials: function (){
        //Game.init();            
        var Ps = [];
        for (var i=0; i<Game.nodesNo; i++){             
            var dist = Game.getDistance(i, Game.Robber);
            Ps.push(dist);              
        }
        Game.debug_alert("Init potentials done");
        return Ps;
    },

    /**
     *  Находит в графе "ловушки"
     */
    initPitfalls: function(){
        Game.pitfalls = [];
        return;
        var M = JSON.parse(JSON.stringify(Game.Matrix));

        for (var i=0; i<M.length; i++){

            for (var k=0; k<M.length; k++)
            	if (Game.getDistanceToCops(k)<getInfinity())
                if (math.sum(M[k])<2){
                    if (Game.pitfalls.indexOf(k)<0)
                    {
                        Game.pitfalls.push(k);                        

                        // Зануляем строку и столбец
                        for (var l=0; l<M.length; l++){
                            M[l][k] = 0;
                            M[k][l] = 0;
                        }
                    }
                }            
        }
    },

    /**
     *  Расчитывает потенциалы вершин
     */
    calcPotetials: function (){
        //Game.init();
        var Ps = [];
        for (var i=0; i<Game.nodesNo; i++){             
            var dist = Game.getDistanceToCops(i);
            Ps.push(dist);              
        }

        return Ps;
        for (var n=1; n<=10; n++)
        {
            var Ps2 = Ps.slice();
            for (var i=0; i<Game.nodesNo; i++){
                
                for (var jk=0; jk<Game.GraphLinkedNodes[i].length; jk++){

                    var j = Game.GraphLinkedNodes[i][jk];
                    
                    if (Game.CopsPos.indexOf(j)>=0)
                        continue;
                    if (Game.Matrix[i][j] >0 ){
                        //Ps2[i] += Ps[j]/(n+1)/(n+1);
                        var Delta = Ps[j]/(n+5)/(n+5);
                        Ps2[i] += Delta;
                        Ps2[j] -= Delta;
                    }

                }


                if (Game.getDistanceToCops(i)==0)
                    Ps2[i] = 0;
                if (Game.pitfalls.indexOf(i)>=0)
                    Ps2[i] = 0;
            }
            Ps = Ps2.slice();
        }

        //Game.debug_alert("Potentials calculated");
        return Ps2;
    }



}



// Примеры графов
if (typeof location !== 'undefined')
{
    Game.GenerateGraph = GraphGenerator.random
    Game.GraphParams.NodesAmount = 5+Math.floor(Math.random()*20)


    if (location.search=="?2"){
        Game.GenerateGraph = GraphGenerator.tree    
    }
    else if (location.search=="?3"){
        Game.GenerateGraph = GraphGenerator.cycle       
        Game.GraphParams.NodesAmount = 2+Math.floor(Math.random()*15)
    }

    else if (location.search=="?4"){
        Game.GenerateGraph = GraphGenerator.grid
        var N = Math.floor(3+Math.random()*5)
        Game.GraphParams.NodesAmount = N*N
    }

    else if (location.search=="?5"){
        Game.GenerateGraph = GraphGenerator.cylinder
        var N = Math.floor(3+Math.random()*5)
        Game.GraphParams.NodesAmount = N*N
    }

    else if (location.search=="?6"){
        Game.GenerateGraph = GraphGenerator.torus
        var N = Math.floor(3+Math.random()*5)
        Game.GraphParams.NodesAmount = N*N
    }

    else if (location.search=="?7"){
        Game.GenerateGraph = GraphGenerator.cube
        var N = Math.floor(2+Math.random()*3)
        Game.GraphParams.NodesAmount = N*N*N
    }

    else if (location.search=="?8"){
        Game.GenerateGraph = GraphGenerator.nCube
        var N = 2;
        Game.GraphParams.NodesAmount = N*N*N*N
        Game.GraphParams.Dimensions = 4;
    }
}